@@ -10,7 +10,10 @@ require 'shellwords'
 require 'rbconfig'
 require 'tempfile'
 require 'tmpdir'
+require 'mcollective/exception'
 require 'mcollective/monkey_patches'
+require "mcollective/translatable"
+require 'mcollective/cache'
 
 # == The Marionette Collective
 #
@@ -20,23 +23,12 @@ require 'mcollective/monkey_patches'
 # and allow you to run agents matching discovery criteria.
 #
 # For an overview of the idea behind this and what it enables please see:
+#
 #   http://www.devco.net/archives/2009/10/18/middleware_for_systems_administration.php
 module MCollective
-  # Exceptions for the RPC system
-  class DDLValidationError<RuntimeError;end
-  class MsgDoesNotMatchRequestID < RuntimeError; end
-  class MsgTTLExpired<RuntimeError;end
-  class NotTargettedAtUs<RuntimeError;end
-  class RPCError<StandardError;end
-  class SecurityValidationFailed<RuntimeError;end
-
-  class InvalidRPCData<RPCError;end
-  class MissingRPCData<RPCError;end
-  class RPCAborted<RPCError;end
-  class UnknownRPCAction<RPCError;end
-  class UnknownRPCError<RPCError;end
-
+  autoload :Agent, "mcollective/agent"
   autoload :Agents, "mcollective/agents"
+  autoload :Aggregate, "mcollective/aggregate"
   autoload :Application, "mcollective/application"
   autoload :Applications, "mcollective/applications"
   autoload :Client, "mcollective/client"
@@ -62,11 +54,12 @@ module MCollective
   autoload :Shell, "mcollective/shell"
   autoload :SSL, "mcollective/ssl"
   autoload :Util, "mcollective/util"
+  autoload :Validator, "mcollective/validator"
   autoload :Vendor, "mcollective/vendor"
 
   MCollective::Vendor.load_vendored
 
-  VERSION="2.1.0"
+  VERSION="2.3.1"
 
   def self.version
     VERSION
